@@ -18,6 +18,8 @@
 
 """
 An implementation of URIs from RFC 3986 (URI Generic Syntax).
+
+See: http://www.ietf.org/rfc/rfc3986.txt
 """
 
 
@@ -25,10 +27,13 @@ from __future__ import unicode_literals
 
 import re
 
+from .kvlist import KeyValueList
+from .util import ustr
+
 
 __all__ = ["general_delimiters", "subcomponent_delimiters",
            "reserved", "unreserved", "percent_encode", "percent_decode",
-           "Authority", "Path", "Query", "URI"]
+           "ParameterString", "Authority", "Path", "Query", "URI"]
 
 
 # RFC 3986 § 2.2.
@@ -42,6 +47,7 @@ unreserved = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"
               "0123456789-._~")
 
 
+# RFC 3986 § 2.1.
 def percent_encode(data, safe=None):
     """ Percent encode a string of data, optionally keeping certain characters
     unencoded.
@@ -51,12 +57,12 @@ def percent_encode(data, safe=None):
         return None
     if isinstance(data, (tuple, list, set)):
         return "&".join(
-            percent_encode(value, safe)
+            percent_encode(value, safe=safe)
             for value in data
         )
     if isinstance(data, dict):
         return "&".join(
-            key + "=" + percent_encode(value, safe)
+            key + "=" + percent_encode(value, safe=safe)
             for key, value in data.items()
         )
     if not safe:
@@ -64,7 +70,7 @@ def percent_encode(data, safe=None):
     try:
         chars = list(data)
     except TypeError:
-        chars = list(str(data))
+        chars = list(ustr(data))
     for i, char in enumerate(chars):
         if char == "%" or (char not in unreserved and char not in safe):
             chars[i] = "".join("%" + hex(b)[2:].upper().zfill(2)
@@ -78,24 +84,38 @@ def percent_decode(data):
     """
     if data is None:
         return None
-    percent_code = re.compile("(%[0-9A-Fa-f]{2})")
+    percent_code = re.compile(r"(%[0-9A-Fa-f]{2}|\+)")
     try:
         bits = percent_code.split(data)
     except TypeError:
-        bits = percent_code.split(str(data))
+        bits = percent_code.split(ustr(data))
     out = bytearray()
     for bit in bits:
         if bit.startswith("%"):
             out.extend(bytearray([int(bit[1:], 16)]))
+        elif bit == "+":
+            out.append(32)
         else:
             out.extend(bytearray(bit, "utf-8"))
     return out.decode("utf-8")
 
 
-class _Part(object):
+class Part(object):
     """ Internal base class for all URI component parts.
     """
 
+    @classmethod
+    def _cast(cls, obj):
+        """ Convert the object supplied to an instance of this class, if
+        possible.
+        """
+        if obj is None:
+            return cls(None)
+        elif isinstance(obj, cls):
+            return obj
+        else:
+            return cls(ustr(obj))
+
     def __init__(self):
         pass
 
@@ -105,6 +125,18 @@ class _Part(object):
     def __str__(self):
         return self.string or ""
 
+    def __eq__(self, other):
+        if other is None:
+            return self.string is None
+        try:
+            other_string = other.string
+        except AttributeError:
+            other_string = self._cast(other).string
+        return self.string == other_string
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
     def __bool__(self):
         return bool(self.string)
 
@@ -112,7 +144,7 @@ class _Part(object):
         return bool(self.string)
 
     def __len__(self):
-        return len(str(self))
+        return len(ustr(self))
 
     def __iter__(self):
         return iter(self.string)
@@ -122,7 +154,82 @@ class _Part(object):
         raise NotImplementedError()
 
 
-class Authority(_Part):
+class ParameterString(Part):
+
+    def __init__(self, string, separator):
+        super(ParameterString, self).__init__()
+        self.__separator = separator
+        self.__none = string is None
+        self.__parameters = KeyValueList()
+        if string:
+            bits = string.split(self.__separator)
+            for bit in bits:
+                if "=" in bit:
+                    key, value = map(percent_decode,
+                                     bit.partition("=")[0::2])
+                else:
+                    key, value = percent_decode(bit), None
+                self.__parameters.append(key, value)
+
+    def __len__(self):
+        return self.__parameters.__len__()
+
+    def __bool__(self):
+        return bool(self.__parameters)
+
+    def __nonzero__(self):
+        return bool(self.__parameters)
+
+    def __contains__(self, item):
+        return self.__parameters.__contains__(item)
+
+    def __hash__(self):
+        return hash(self.string)
+
+    def __iter__(self):
+        return self.__parameters.__iter__()
+
+    def __getitem__(self, index):
+        if isinstance(index, slice):
+            out = ParameterString("", self.__separator)
+            out.__parameters.extend(self.__parameters.__getitem__(index))
+            return out
+        else:
+            return self.__parameters.__getitem__(index)
+
+    def __getslice__(self, start, stop):
+        out = ParameterString("", self.__separator)
+        out.__parameters.extend(self.__parameters.__getslice__(start, stop))
+        return out
+
+    def get(self, name, index=0):
+        if not self.__parameters.has_key(name):
+            raise KeyError(name)
+        for i, value in enumerate(self.__parameters.get(name)):
+            if i == index:
+                return value
+        raise IndexError("Parameter {0} does not have {1} "
+                         "values".format(name, index))
+
+    def get_all(self, name):
+        if not self.__parameters.has_key(name):
+            raise KeyError(name)
+        return list(self.__parameters.get(name))
+
+    @property
+    def string(self):
+        if self.__none:
+            return None
+        bits = []
+        for key, value in self.__parameters:
+            if value is None:
+                bits.append(percent_encode(key))
+            else:
+                bits.append(percent_encode(key) + "=" + percent_encode(value))
+        return self.__separator.join(bits)
+
+
+class Authority(Part):
     """ A host name plus optional port and user information detail.
 
     **Syntax**
@@ -135,13 +242,14 @@ class Authority(_Part):
     """
 
     @classmethod
-    def __cast(cls, obj):
-        if obj is None:
-            return cls(None)
-        elif isinstance(obj, cls):
-            return obj
+    def _parse_host_port(cls, string):
+        if ":" in string:
+            host, port = string.rpartition(":")[0::2]
+            port = int(port)
         else:
-            return cls(str(obj))
+            host = string
+            port = None
+        return host, port
 
     def __init__(self, string):
         super(Authority, self).__init__()
@@ -155,24 +263,7 @@ class Authority(_Part):
                 self.__user_info = percent_decode(self.__user_info)
             else:
                 self.__user_info = None
-            if ":" in string:
-                self.__host, self.__port = string.rpartition(":")[0::2]
-                self.__port = int(self.__port)
-            else:
-                self.__host = string
-                self.__port = None
-
-    def __eq__(self, other):
-        other = self.__cast(other)
-        return (self.__user_info == other.__user_info and
-                self.__host == other.__host and
-                self.__port == other.__port)
-
-    def __ne__(self, other):
-        other = self.__cast(other)
-        return (self.__user_info != other.__user_info or
-                self.__host != other.__host or
-                self.__port != other.__port)
+            self.__host, self.__port = self._parse_host_port(string)
 
     def __hash__(self):
         return hash(self.string)
@@ -225,7 +316,7 @@ class Authority(_Part):
         """
         u = [self.__host]
         if self.__port is not None:
-            u += [":", str(self.__port)]
+            u += [":", ustr(self.__port)]
         return "".join(u)
 
     @property
@@ -281,7 +372,7 @@ class Authority(_Part):
             u += [percent_encode(self.__user_info), "@"]
         u += [self.__host]
         if self.__port is not None:
-            u += [":", str(self.__port)]
+            u += [":", ustr(self.__port)]
         return "".join(u)
         
     @property
@@ -309,16 +400,7 @@ class Authority(_Part):
         return self.__user_info
 
 
-class Path(_Part):
-
-    @classmethod
-    def __cast(cls, obj):
-        if obj is None:
-            return cls(None)
-        elif isinstance(obj, cls):
-            return obj
-        else:
-            return cls(str(obj))
+class Path(Part):
 
     def __init__(self, string):
         super(Path, self).__init__()
@@ -327,14 +409,6 @@ class Path(_Part):
         else:
             self.__segments = list(map(percent_decode, string.split("/")))
 
-    def __eq__(self, other):
-        other = self.__cast(other)
-        return self.__segments == other.__segments
-
-    def __ne__(self, other):
-        other = self.__cast(other)
-        return self.__segments != other.__segments
-
     def __hash__(self):
         return hash(self.string)
 
@@ -404,92 +478,18 @@ class Path(_Part):
             return self
 
 
-class Query(_Part):
+class Query(ParameterString):
 
-    @classmethod
-    def __cast(cls, obj):
-        if obj is None:
-            return cls(None)
-        elif isinstance(obj, cls):
-            return obj
-        else:
-            return cls(str(obj))
-
-    @classmethod
-    def encode(cls, iterable):
-        if iterable is None:
-            return None
-        bits = []
-        if isinstance(iterable, dict):
-            for key, value in iterable.items():
-                if value is None:
-                    bits.append(percent_encode(key))
-                else:
-                    bits.append(percent_encode(key) + "=" + percent_encode(value))
-        else:
-            for item in iterable:
-                if isinstance(item, tuple) and len(item) == 2:
-                    key, value = item
-                    if value is None:
-                        bits.append(percent_encode(key))
-                    else:
-                        bits.append(percent_encode(key) + "=" + percent_encode(value))
-                else:
-                    bits.append(percent_encode(item))
-        return "&".join(bits)
-
-    @classmethod
-    def decode(cls, string):
-        if string is None:
-            return None
-        data = []
-        if string:
-            bits = string.split("&")
-            for bit in bits:
-                if "=" in bit:
-                    key, value = map(percent_decode, bit.partition("=")[0::2])
-                else:
-                    key, value = percent_decode(bit), None
-                data.append((key, value))
-        return data
+    SEPARATOR = "&"
 
     def __init__(self, string):
-        super(Query, self).__init__()
-        self.__query = Query.decode(string)
-        if self.__query is None:
-            self.__query_dict = None
-        else:
-            self.__query_dict = dict(self.__query)
-
-    def __eq__(self, other):
-        other = self.__cast(other)
-        return self.__query == other.__query
-
-    def __ne__(self, other):
-        other = self.__cast(other)
-        return self.__query != other.__query
+        super(Query, self).__init__(string, self.SEPARATOR)
 
     def __hash__(self):
         return hash(self.string)
 
-    @property
-    def string(self):
-        return Query.encode(self.__query)
 
-    def __iter__(self):
-        if self.__query is None:
-            return iter(())
-        else:
-            return iter(self.__query)
-
-    def __getitem__(self, key):
-        if self.__query_dict is None:
-            raise KeyError(key)
-        else:
-            return self.__query_dict[key]
-
-
-class URI(_Part):
+class URI(Part):
     """ Uniform Resource Identifier.
 
     .. seealso::
@@ -499,37 +499,90 @@ class URI(_Part):
     """
 
     @classmethod
-    def __cast(cls, obj):
-        if obj is None:
-            return cls(None)
-        elif isinstance(obj, cls):
-            return obj
+    def build(cls, **parts):
+        """ Build a URI object from named parts. The part names available are:
+
+        - string
+        - hierarchical_part
+        - absolute_path_reference
+        - authority
+        - host_port
+        - scheme
+        - user_info
+        - host
+        - port
+        - path
+        - query
+        - fragment
+
+        """
+        uri = URI(parts.get("string"))
+        uri.__set_hierarchical_part(parts.get("hierarchical_part"))
+        uri.__set_absolute_path_reference(parts.get("absolute_path_reference"))
+        uri.__set_authority(parts.get("authority"))
+        uri.__set_host_port(parts.get("host_port"))
+        uri.__set_scheme(parts.get("scheme"))
+        uri.__set_user_info(parts.get("user_info"))
+        uri.__set_host(parts.get("host"))
+        uri.__set_port(parts.get("port"))
+        uri.__set_path(parts.get("path"))
+        uri.__set_query(parts.get("query"))
+        uri.__set_fragment(parts.get("fragment"))
+        if parts and uri.__path is None:
+            uri.__path = Path("")
+        return uri
+
+    @classmethod
+    def _partition_fragment(cls, value):
+        if "#" in value:
+            value, fragment = value.partition("#")[0::2]
+            fragment = percent_decode(fragment)
         else:
-            return cls(str(obj))
+            fragment = None
+        return value, fragment
+
+    @classmethod
+    def _partition_query(cls, value):
+        if "?" in value:
+            value, query = value.partition("?")[0::2]
+            query = Query(query)
+        else:
+            query = None
+        return value, query
+
+    @classmethod
+    def _parse_hierarchical_part(cls, value):
+        if value.startswith("//"):
+            value = value[2:]
+            slash = value.find("/")
+            if slash >= 0:
+                authority = Authority(value[:slash])
+                path = Path(value[slash:])
+            else:
+                authority = Authority(value)
+                path = Path("")
+        else:
+            authority = None
+            path = Path(value)
+        return authority, path
 
     def __init__(self, value):
         super(URI, self).__init__()
+        self.__scheme = None
+        self.__authority = None
+        self.__path = None
+        self.__query = None
+        self.__fragment = None
         try:
             if value.__uri__ is None:
-                self.__scheme = None
-                self.__authority = None
-                self.__path = None
-                self.__query = None
-                self.__fragment = None
                 return
         except AttributeError:
             pass
-        if value is None:
-            self.__scheme = None
-            self.__authority = None
-            self.__path = None
-            self.__query = None
-            self.__fragment = None
-        else:
+        if value is not None:
             try:
-                value = str(value.__uri__)
+                value = ustr(value.__uri__)
             except AttributeError:
-                value = str(value)
+                value = ustr(value)
             # scheme
             if ":" in value:
                 self.__scheme, value = value.partition(":")[0::2]
@@ -537,46 +590,11 @@ class URI(_Part):
             else:
                 self.__scheme = None
             # fragment
-            if "#" in value:
-                value, self.__fragment = value.partition("#")[0::2]
-                self.__fragment = percent_decode(self.__fragment)
-            else:
-                self.__fragment = None
+            value, self.__fragment = self._partition_fragment(value)
             # query
-            if "?" in value:
-                value, self.__query = value.partition("?")[0::2]
-                self.__query = Query(self.__query)
-            else:
-                self.__query = None
+            value, self.__query = self._partition_query(value)
             # hierarchical part
-            if value.startswith("//"):
-                value = value[2:]
-                slash = value.find("/")
-                if slash >= 0:
-                    self.__authority = Authority(value[:slash])
-                    self.__path = Path(value[slash:])
-                else:
-                    self.__authority = Authority(value)
-                    self.__path = Path("")
-            else:
-                self.__authority = None
-                self.__path = Path(value)
-
-    def __eq__(self, other):
-        other = self.__cast(other)
-        return (self.__scheme == other.__scheme and
-                self.__authority == other.__authority and
-                self.__path == other.__path and
-                self.__query == other.__query and
-                self.__fragment == other.__fragment)
-
-    def __ne__(self, other):
-        other = self.__cast(other)
-        return (self.__scheme != other.__scheme or
-                self.__authority != other.__authority or
-                self.__path != other.__path or
-                self.__query != other.__query or
-                self.__fragment != other.__fragment)
+            self.__authority, self.__path = self._parse_hierarchical_part(value)
 
     def __hash__(self):
         return hash(self.string)
@@ -585,6 +603,64 @@ class URI(_Part):
     def __uri__(self):
         return self.string
 
+    def __set_hierarchical_part(self, string):
+        if string is not None:
+            self.__authority, self.__path = self._parse_hierarchical_part(string)
+
+    def __set_absolute_path_reference(self, string):
+        if string is not None:
+            string, self.__fragment = self._partition_fragment(string)
+            string, self.__query = self._partition_query(string)
+            self.__path = Path(string)
+
+    def __set_authority(self, string):
+        if string is not None:
+            self.__authority = Authority(string)
+
+    def __set_host_port(self, string):
+        if string is not None:
+            if self.__authority is None:
+                self.__authority = Authority(string)
+            else:
+                host, port = Authority._parse_host_port(string)
+                self.__authority._Authority__host = host
+                self.__authority._Authority__port = port
+
+    def __set_scheme(self, string):
+        if string is not None:
+            self.__scheme = string
+
+    def __set_user_info(self, string):
+        if string is not None:
+            if self.__authority is None:
+                self.__authority = Authority("")
+            self.__authority._Authority__user_info = string
+
+    def __set_host(self, string):
+        if string is not None:
+            if self.__authority is None:
+                self.__authority = Authority(string)
+            else:
+                self.__authority._Authority__host = string
+
+    def __set_port(self, number):
+        if number is not None:
+            if self.__authority is None:
+                self.__authority = Authority("")
+            self.__authority._Authority__port = number
+
+    def __set_path(self, string):
+        if string is not None:
+            self.__path = Path(string)
+
+    def __set_query(self, string):
+        if string is not None:
+            self.__query = Query(string)
+
+    def __set_fragment(self, string):
+        if string is not None:
+            self.__fragment = string
+
     @property
     def string(self):
         """ The full percent-encoded string value of this URI or
@@ -624,10 +700,10 @@ class URI(_Part):
         if self.__scheme is not None:
             u += [percent_encode(self.__scheme), ":"]
         if self.__authority is not None:
-            u += ["//", str(self.__authority)]
-        u += [str(self.__path)]
+            u += ["//", ustr(self.__authority)]
+        u += [ustr(self.__path)]
         if self.__query is not None:
-            u += ["?", str(self.__query)]
+            u += ["?", ustr(self.__query)]
         if self.__fragment is not None:
             u += ["#", percent_encode(self.__fragment)]
         return "".join(u)
@@ -855,8 +931,8 @@ class URI(_Part):
             return None
         u = []
         if self.__authority is not None:
-            u += ["//", str(self.__authority)]
-        u += [str(self.__path)]
+            u += ["//", ustr(self.__authority)]
+        u += [ustr(self.__path)]
         return "".join(u)
 
     @property
@@ -878,20 +954,20 @@ class URI(_Part):
         """
         if self.__path is None:
             return None
-        u = [str(self.__path)]
+        u = [ustr(self.__path)]
         if self.__query is not None:
-            u += ["?", str(self.__query)]
+            u += ["?", ustr(self.__query)]
         if self.__fragment is not None:
             u += ["#", percent_encode(self.__fragment)]
         return "".join(u)
 
     def _merge_path(self, relative_path_reference):
         if self.__authority is not None and not self.__path:
-            return Path("/" + str(relative_path_reference))
+            return Path("/" + ustr(relative_path_reference))
         elif "/" in self.__path.string:
             segments = self.__path.segments
             segments[-1] = ""
-            return Path("/".join(segments) + str(relative_path_reference))
+            return Path("/".join(segments) + ustr(relative_path_reference))
         else:
             return relative_path_reference
 
@@ -906,7 +982,7 @@ class URI(_Part):
         """
         if reference is None:
             return None
-        reference = self.__cast(reference)
+        reference = self._cast(reference)
         target = URI(None)
         if not strict and reference.__scheme == self.__scheme:
             reference_scheme = None
@@ -930,7 +1006,7 @@ class URI(_Part):
                     else:
                         target.__query = self.__query
                 else:
-                    if str(reference.__path).startswith("/"):
+                    if ustr(reference.__path).startswith("/"):
                         target.__path = reference.__path.remove_dot_segments()
                     else:
                         target.__path = self._merge_path(reference.__path)
